from typing import Dict, List, Union
from dataclasses import dataclass

# Add alias to everything
from queenbee.io.outputs.alias import (
    DAGGenericOutputAlias, DAGStringOutputAlias, DAGIntegerOutputAlias,
    DAGNumberOutputAlias, DAGBooleanOutputAlias, DAGFolderOutputAlias,
    DAGFileOutputAlias, DAGPathOutputAlias, DAGJSONObjectOutputAlias,
    DAGArrayOutputAlias, DAGLinkedOutputAlias
)
from queenbee.base.basemodel import BaseModel
from queenbee.io.common import ItemType, IOAliasHandler

from pydantic import Field


__all__ = ('Outputs', )

_outputs_mapper = {
    'GenericOutputAlias': DAGGenericOutputAlias,
    'StringOutputAlias': DAGStringOutputAlias,
    'IntegerOutputAlias': DAGIntegerOutputAlias,
    'NumberOutputAlias': DAGNumberOutputAlias,
    'BooleanOutputAlias': DAGBooleanOutputAlias,
    'FolderOutputAlias': DAGFolderOutputAlias,
    'FileOutputAlias': DAGFileOutputAlias,
    'PathOutputAlias': DAGPathOutputAlias,
    'DictOutputAlias': DAGJSONObjectOutputAlias,
    'ListOutputAlias': DAGArrayOutputAlias,
    'LinkedOutputAlias': DAGLinkedOutputAlias
}


def _get_from(value, reference):
    try:
        # only TaskReference has parent
        parent = value['parent']
    except TypeError:
        # value
        return {'type': reference, 'path': value}
    else:
        # Task reference values and names are nested
        variable = value['name']
        value = value['value']
        return {'type': reference, 'name': parent, 'variable': variable}


class _OutputAliasBase(BaseModel):
    name: str = Field(...)
    annotations: Dict = None
    description: str = None
    platform: List[str] = Field(
        ...,
        description='Name of the client platform (e.g. Grasshopper, Revit, etc). The '
        'value can be any strings as long as it has been agreed between client-side '
        'developer and author of the recipe.'
    )
    handler: List[IOAliasHandler] = Field(
        ...,
        description='List of process actions to process the input or output value.'
    )

    def to_queenbee(self):
        """Convert this output to a Queenbee DAG output."""
        func = _outputs_mapper[self.__class__.__name__]
        data = {
            'name': self.name,
            'description': self.description,
            'annotations': self.annotations,
            'platform': self.platform,
            'handler': [h.dict() for h in self.handler]
        }

        if hasattr(self, 'items_type'):
            data['items_type'] = self.items_type

        return func.parse_obj(data)

    @property
    def is_artifact(self):
        return False


class GenericOutputAlias(_OutputAliasBase):
    """ A DAG generic output alias.

    Args:
        name: Alias input name.
        annotations: An optional annotation dictionary.
        description: Input description.
        platform: A list of names for the client platform (e.g. Grasshopper, Revit, etc).
            The value can be any strings as long as it has been agreed between
            client-side developer and author of the recipe.
        handler: List of process actions to process the input or output value.

    """
    ...


class LinkedOutputAlias(GenericOutputAlias):
    """ A DAG linked output alias.

    Args:
        name: Alias input name.
        annotations: An optional annotation dictionary.
        description: Input description.
        platform: A list of names for the client platform (e.g. Grasshopper, Revit, etc).
            The value can be any strings as long as it has been agreed between
            client-side developer and author of the recipe.
        handler: List of process actions to process the input or output value.

    """
    ...


class StringOutputAlias(GenericOutputAlias):
    """ A DAG string output alias.

    Args:
        name: Alias input name.
        annotations: An optional annotation dictionary.
        description: Input description.
        platform: A list of names for the client platform (e.g. Grasshopper, Revit, etc).
            The value can be any strings as long as it has been agreed between
            client-side developer and author of the recipe.
        handler: List of process actions to process the input or output value.

    """
    ...


class IntegerOutputAlias(StringOutputAlias):
    """ A DAG integer output alias.

    Args:
        name: Alias input name.
        annotations: An optional annotation dictionary.
        description: Input description.
        platform: A list of names for the client platform (e.g. Grasshopper, Revit, etc).
            The value can be any strings as long as it has been agreed between
            client-side developer and author of the recipe.
        handler: List of process actions to process the input or output value.

    """
    ...


class NumberOutputAlias(StringOutputAlias):
    """ A DAG number output alias.

    Args:
        name: Alias input name.
        annotations: An optional annotation dictionary.
        description: Input description.
        platform: A list of names for the client platform (e.g. Grasshopper, Revit, etc).
            The value can be any strings as long as it has been agreed between
            client-side developer and author of the recipe.
        handler: List of process actions to process the input or output value.

    """
    ...


class BooleanOutputAlias(StringOutputAlias):
    """ A DAG boolean output alias.

    Args:
        name: Alias input name.
        annotations: An optional annotation dictionary.
        description: Input description.
        platform: A list of names for the client platform (e.g. Grasshopper, Revit, etc).
            The value can be any strings as long as it has been agreed between
            client-side developer and author of the recipe.
        handler: List of process actions to process the input or output value.

    """
    ...


class DictOutputAlias(StringOutputAlias):
    """ A DAG dictionary output alias.

    Args:
        name: Alias input name.
        annotations: An optional annotation dictionary.
        description: Input description.
        platform: A list of names for the client platform (e.g. Grasshopper, Revit, etc).
            The value can be any strings as long as it has been agreed between
            client-side developer and author of the recipe.
        handler: List of process actions to process the input or output value.

    """
    ...


class ListOutputAlias(StringOutputAlias):
    """ A DAG list output alias.

    Args:
        name: Alias input name.
        annotations: An optional annotation dictionary.
        description: Input description.
        platform: A list of names for the client platform (e.g. Grasshopper, Revit, etc).
            The value can be any strings as long as it has been agreed between
            client-side developer and author of the recipe.
        handler: List of process actions to process the input or output value.

    """
    items_type: ItemType = Field(
        ItemType.String,
        description='Type of items in this array. All the items in an array must be '
        'from the same type.'
    )


class FolderOutputAlias(StringOutputAlias):
    """ A DAG folder output alias.

    Args:
        name: Alias input name.
        annotations: An optional annotation dictionary.
        description: Input description.
        platform: A list of names for the client platform (e.g. Grasshopper, Revit, etc).
            The value can be any strings as long as it has been agreed between
            client-side developer and author of the recipe.
        handler: List of process actions to process the input or output value.

    """

    @property
    def is_artifact(self):
        return True


class FileOutputAlias(FolderOutputAlias):
    """ A DAG file output alias.

    Args:
        name: Alias input name.
        annotations: An optional annotation dictionary.
        description: Input description.
        platform: A list of names for the client platform (e.g. Grasshopper, Revit, etc).
            The value can be any strings as long as it has been agreed between
            client-side developer and author of the recipe.
        handler: List of process actions to process the input or output value.

    """
    ...


class PathOutputAlias(FolderOutputAlias):
    """ A DAG path output alias. A path can be a file or a folder.

    Args:
        name: Alias input name.
        annotations: An optional annotation dictionary.
        description: Input description.
        platform: A list of names for the client platform (e.g. Grasshopper, Revit, etc).
            The value can be any strings as long as it has been agreed between
            client-side developer and author of the recipe.
        handler: List of process actions to process the input or output value.

    """
    ...


@dataclass
class Outputs:
    """DAG outputs enumeration."""
    any = GenericOutputAlias
    str = StringOutputAlias
    int = IntegerOutputAlias
    float = NumberOutputAlias
    bool = BooleanOutputAlias
    file = FileOutputAlias
    folder = FolderOutputAlias
    path = PathOutputAlias
    dict = DictOutputAlias
    list = ListOutputAlias
    link = LinkedOutputAlias


OutputAliasTypes = Union[
    GenericOutputAlias, StringOutputAlias, IntegerOutputAlias, NumberOutputAlias,
    BooleanOutputAlias, FileOutputAlias, FolderOutputAlias, PathOutputAlias,
    DictOutputAlias, ListOutputAlias, LinkedOutputAlias
]



"""Pollination DSL command line interface."""
import pathlib
import os
import sys
import tempfile

import click
from click.exceptions import ClickException
from queenbee.cli.context import Context as QueenbeeContext
from queenbee.plugin.plugin import Plugin
import yaml

from queenbee_pollination.cli.push import recipe, plugin
from queenbee_local.cli import run_recipe
from pollination_dsl.package import translate, load, _init_repo
from pollination_dsl.common import _get_package_readme, _get_package_owner, \
    name_to_pollination


@click.group()
@click.version_option()
def main():
    pass


@click.group(help='interact with pollination python DSL packages.')
@click.version_option()
@click.pass_context
def dsl(ctx):
    """Pollination python DSL plugin."""
    from queenbee_pollination.cli.context import Context
    ctx.obj = Context()


@dsl.command('translate')
@click.argument('recipe_name')
@click.argument(
    'target-folder',
    type=click.Path(exists=False, file_okay=False, resolve_path=True, dir_okay='True'),
    default='.'
)
@click.option(
    '--queenbee/--luigi', is_flag=True, default=True, help='Switch between a baked '
    'Queenbee recipe or a luigi pipeline. To translate to a luigi pipeline you must '
    'have queenbee-luigi package installed.'
)
@click.pass_context
def translate_recipe(ctx, recipe_name, target_folder, queenbee, no_exit=False):
    """Translate a Pollination recipe to a Queenbee recipe or a Luigi pipeline.

    You can use queenbee local run command to run the pipline after export.

    \b
    Args:\b
        recipe-name: Recipe name. Recipe must be installed as a Python package.\b
        target-folder: Path to target folder to translate the recipe.\b

    """
    folder = pathlib.Path(target_folder)
    folder.mkdir(exist_ok=True)
    if not recipe_name.startswith('pollination'):
        recipe_name = f'pollination-{recipe_name}'

    if queenbee:
        recipe_folder = translate(recipe_name, target_folder=target_folder, baked=True)
        print(f'Success: {recipe_folder}', file=sys.stderr)
        if no_exit:
            return
        return sys.exit(0)

    else:
        try:
            from queenbee_luigi.recipe import Recipe
        except ImportError:
            raise ClickException(
                'Failed to find queenbee-luigi. To translate a recipe to luigi pipeline '
                'you must have queenbee-luigi installed.'
            )

        recipe = load(recipe_name, baked=True)
        try:
            rep = Recipe(recipe)
            recipe_folder = rep.write(target_folder=target_folder)
        except Exception as e:
            print(f'Failed to translate the recipe:\n{e}', file=sys.stderr)
            return sys.exit(1)
        else:
            print(f'Success: {recipe_folder}', file=sys.stderr)
            if no_exit:
                return
            return sys.exit(0)


@dsl.command('push')
@click.argument('package_name')
@click.option(
    '-e', '--endpoint', help='Endpoint to push the resource.', show_default=True,
    default='https://api.pollination.cloud'
)
# TODO: Add better support for mapping different dependencies to different sources. For
# now it is all set to the same value which is fine for our use cases.
@click.option(
    '-src', '--source', help='A link to replace the source for dependencies. This value '
    'will overwrite the source value in recipe\'s dependencies files. By default it '
    'will be set to https://api.pollination.cloud/registries'
)
@click.option(
    '--public/--private', help='Indicate if the recipe or plugin should be created as '
    'a public or a private resource. This option does not change the visibility of a '
    'resource if it already exist.', is_flag=True, default=True
)
@click.option(
    '--tag', '-t', help='An optional tag to enforce tag number. By default the tag will '
    'be extracted from package version.',
)
@click.option(
    '--dry-run', '-dr', help='An option to test the command and export the package to '
    'a folder without pushing it to endpoint.', is_flag=True, default=False
)
@click.pass_context
def push_resource(ctx, package_name, endpoint, source, public, tag, dry_run):
    """Push a pollination dsl recipe or plugin to Pollination.

    To run this command you need queenbee-pollination[cli] installed. You can also
    get these libraries by running ``pip install pollination-dsl[pollination]`` command.

    \b
    Args:
        package_name: The name of the recipe or plugin. Recipe must be installed as a
            Python package.

    """
    # set the config vars
    ctx.obj.config.endpoint = endpoint
    ctx.obj.config.token = os.getenv('QB_POLLINATION_TOKEN')
    if not dry_run:
        assert ctx.obj.config.token is not None, \
            'Pollination token is not set. Use QB_POLLINATION_TOKEN to set it as an ' \
            'env variable. You can generate an API key under settings from your ' \
            'Pollination profile.'

    if ctx.obj.queenbee is None:
        ctx.obj.queenbee = QueenbeeContext()

    resource = load(package_name, False)

    # get package owner
    owner = _get_package_owner(package_name)

    if isinstance(resource, Plugin):
        cmd = plugin
        resource_type = 'plugin'
    else:
        cmd = recipe
        resource_type = 'recipe'

    sub_folder = package_name.replace('_', '-').replace('pollination-', '')

    # write to a folder
    temp_dir = tempfile.mkdtemp()
    folder = pathlib.Path(temp_dir, sub_folder)

    resource.to_folder(
        folder_path=folder,
        readme_string=_get_package_readme(package_name)
    )

    if tag is None:
        tag = resource.metadata.tag

    # overwite resources in dependencies
    if resource_type == 'recipe':
        source = source or 'https://api.pollination.cloud/registries'
        # update the value for source in dependencies.yaml file
        dep_file = pathlib.Path(folder, 'dependencies.yaml')
        data = yaml.safe_load(dep_file.read_bytes())
        for dep in data['dependencies']:
            dep_name = name_to_pollination(dep['name'])
            owner = _get_package_owner(dep_name)
            dep['source'] = pathlib.Path(source, owner).as_posix()
        yaml.dump(data, dep_file.open('w'))

    if dry_run:
        print(folder, file=sys.stderr)
    else:
        ctx.invoke(
            cmd, path=folder, owner=owner, tag=tag, create_repo=True, public=public
        )


@dsl.command('run')
@click.argument('recipe_name')
@click.argument(
    'project_folder',
    type=click.Path(exists=True, file_okay=False, resolve_path=True, dir_okay=True),
    default='.'
)
@click.option(
    '-i', '--inputs',
    type=click.Path(exists=True, file_okay=True, resolve_path=True, dir_okay=False),
    default=None, show_default=True, help='Path to the JSON file to'
    ' overwrite inputs for this recipe.'
)
@click.option(
    '-w', '--workers', type=int, default=1, show_default=True,
    help='Number of workers to execute tasks in parallel.'
)
@click.option(
    '-e', '--env', multiple=True, help='An option to pass environmental variables to '
    'commands. Use = to separate key and value. RAYPATH=/usr/local/lib/ray'
)
@click.option('-n', '--name', help='Simulation name for this run.')
@click.option(
    '-d', '--debug',
    type=click.Path(exists=False, file_okay=False, resolve_path=True, dir_okay=True),
    help='Optional path to a debug folder. If debug folder is provided all the steps '
    'of the simulation will be executed inside the debug folder which can be used for '
    'furthur inspection.'
)
@click.pass_context
def run(ctx, recipe_name, project_folder, inputs, workers, env, name, debug):
    """Execute a recipe against a project folder.

    \b
    Args:
        recipe_name: Name of the recipe (e.g. pollination-daylight-factor)
        project_folder: Path to project folder. Project folder includes the input files
            and folders for the recipe.

    """
    # translate the recipe if it hasn't been translated already
    target_folder = _init_repo().parent / 'pollination-luigi'
    # TODO: Check for the recipe from the same version and don't generate it every
    # time. Add an overwrite option to the command.
    ctx.invoke(
        translate_recipe,
        recipe_name=recipe_name,
        target_folder=target_folder.as_posix(),
        queenbee=False,
        no_exit=True
    )
    # get clean recipe name
    recipe_name = recipe_name.replace('-', '_').replace('pollination_', '') \
        .replace('pollination.', '')

    recipe_folder = target_folder / recipe_name

    # run the recipe using queenbee-local
    ctx.invoke(
        run_recipe, recipe=recipe_folder, project_folder=project_folder, inputs=inputs,
        workers=workers, env=env, name=name, debug=debug
    )


main.add_command(dsl)



from dataclasses import dataclass
from typing import NamedTuple, Dict, List
import importlib
from collections import namedtuple
import importlib_metadata
import re
import pkg_resources
import warnings
import subprocess
import sys

from queenbee.plugin.plugin import MetaData


def camel_to_snake(name: str) -> str:
    """Change name from CamelCase to snake-case."""
    return name[0].lower() + \
        ''.join(['-' + x.lower() if x.isupper() else x for x in name][1:])


def name_to_pollination(name: str) -> str:
    """Add a pollination- in front of the name."""
    if name.startswith('pollination.'):
        return name.replace('pollination.', 'pollination-')
    elif name.replace('_', '-').startswith('pollination-'):
        return name
    else:
        return f'pollination-{name}'


def import_module(name: str, pull=True):
    """Import a module by name.

    This function only works for pollination namespace packages.
    If the module is not installed and pull is set to True it will try to pull the
    package from PyPI.

    """
    package_name = name_to_pollination(name).replace('-', '_')
    err_msg = \
        f'Failed to import \'{package_name}\' from local installation. ' \
        'Trying to install from PyPI.'

    # for pollination modules split and try again pollination-honeybee-radiance
    # is namedspaced as pollination.honeybee_radiance
    package_name_segments = package_name.split('_')
    _namespace = package_name_segments[0]
    _name = '_'.join(package_name_segments[1:])
    try:
        namespace = __import__(f'{_namespace}.{_name}')
        module = getattr(namespace, _name)
    except ModuleNotFoundError:
        if pull:
            print(err_msg)
            success = _try_pull_from_pip(package=package_name)
            if success:
                # try again
                return import_module(package_name)
        raise ModuleNotFoundError(
            f'Failed to import \'{package_name}\' locally or from PyPI.'
        )

    return module


def _try_pull_from_pip(package):
    try:
        subprocess.check_call(
            [sys.executable, '-m', 'pip', 'install', package]
        )
        return True
    except subprocess.CalledProcessError:
        # it failed to download from PyPI
        # error message will be printed out by pip
        return False


def get_requirement_version(package_name, dependency_name):
    """Get assigned version to a dependency in package requirements."""
    fixed_name = package_name.replace('pollination.', 'pollination_')
    dependency_name = dependency_name.replace('_', '-')
    requirements = {}
    try:
        req = pkg_resources.get_distribution(fixed_name).get_metadata('requires.txt')
    except FileNotFoundError:
        # try to get it from meta data
        package_data = importlib_metadata.metadata(fixed_name)
        req_dists = package_data.get_all('Requires-Dist') or []
        for package in req_dists:
            name, version = package.split(' (')
            version = \
                version.replace('=', '').replace('>', '').replace('<', '') \
                .replace(')', '').strip()
            requirements[name] = version
    else:
        for package in pkg_resources.parse_requirements(req):
            version = \
                str(package.specifier).replace('=', '').replace('>', '').replace('<', '')
            requirements[package.project_name] = version

    assert dependency_name in requirements, \
        f'{dependency_name} is not a requirement for {package_name}.'

    return requirements[dependency_name]


def get_docker_image_from_dependency(package, dependency, owner):
    """Get a docker image id from package information.

    Arguments:
        package: Name of the package (e.g. pollination-honeybee-radiance)
        dependency: Name of the dependency (e.g. honeybee-radiance)
        owner: The account owner for docker image

    Returns:
        str -- docker image id (e.g. ladybugtools/honeybee-radiance:0.5.2)
    """
    try:
        image_version = get_requirement_version(package, dependency)
        image_id = f'ladybugtools/{dependency}:{image_version}'
    except (FileNotFoundError, AssertionError) as error:
        # this should not happen if the package is installed correctly
        # but Python has so many ways to store requirements based on how the package
        # is built and where! It's better to set it to latest instead of failing.
        warnings.warn(
            f'Failed to pinpoint the version for {dependency} as a dependency for'
            f' {package}. Will set the docker version to latest.\n{error}'
        )
        image_id = f'ladybugtools/{dependency}:latest'

    return image_id


def _get_package_readme(package_name: str) -> str:
    package_data = importlib_metadata.metadata(package_name.replace('-', '_'))
    long_description = package_data.get_payload()
    if not long_description.strip():
        content = package_data.get('Description')
        long_description = []
        for line in content.split('\n'):
            long_description.append(re.sub('^        ', '', line))
        long_description = '\n'.join(long_description)
    return long_description


def _get_package_owner(package_name: str) -> str:
    """Author field is used for package owner."""
    package_data = importlib_metadata.metadata(package_name.replace('-', '_'))
    owner = package_data.get('Author')
    assert owner, \
        'You must set the author of the package in setup.py to Pollination account owner'
    # ensure there is only one author
    owner = owner.strip()
    assert len(owner.split(',')) == 1, \
        'A Pollination package can only have one author. Use maintainer field for ' \
        'providing multiple maintainers.'

    return owner


def _get_package_license(package_data: Dict) -> Dict:
    # try to get license
    license_info = package_data.get('License')
    if not license_info:
        license, link = None, None
    elif license_info and ',' in license_info:
        license, link = [info.strip() for info in license_info.split(',')]
    else:
        license, link = license_info.strip(), None

    return {'name': license, 'url': link}


def _get_package_keywords(package_data: Dict) -> List:
    keywords = package_data.get('Keyword')
    if keywords:
        keywords = [key.strip() for key in keywords.split(',')]
    return keywords


def _get_package_icon(package_data: Dict) -> str:
    urls = package_data.get_all('Project-URL')
    for url in urls:
        key, value = url.split(',')
        if key == 'icon':
            return value.strip()


def _get_package_maintainers(package_data: Dict) -> List[Dict]:
    package_maintainers = []
    maintainer = package_data.get('Maintainer')
    maintainer_email = package_data.get('Maintainer-email')

    if maintainer:
        maintainers = [m.strip() for m in maintainer.split(',')]
        if maintainer_email:
            emails = [m.strip() for m in maintainer_email.split(',')]
            for name, email in zip(maintainers, emails):
                package_maintainers.append({'name': name, 'email': email})
        else:
            for name in maintainers:
                package_maintainers.append({'name': name, 'email': None})
    return package_maintainers


def _get_package_version(package_data: Dict) -> str:
    """Get package version.

    This function returns the non-development version for a development version.
    It removes the .dev part and return x.y.z-1 version if it is a dev version.
    """
    version = package_data.get('Version')
    xyz = version.split('.')
    if len(xyz) <= 3:
        return version
    # clean up the developer version
    x, y, z = xyz[:3]
    return f'{x}.{y}.{int(z) - 1}'


def _get_package_data(package_name: str) -> Dict:

    package_data = importlib_metadata.metadata(package_name)

    data = {
        'name': package_data.get('Name').replace('pollination-', ''),
        'description': package_data.get('Summary'),
        'home': package_data.get('Home-page'),
        'tag': _get_package_version(package_data),
        'keywords': _get_package_keywords(package_data),
        'maintainers': _get_package_maintainers(package_data),
        'license': _get_package_license(package_data),
        'icon': _get_package_icon(package_data)
    }

    return data


def _get_meta_data(module, package_type: str) -> MetaData:
    qb_info = module.__pollination__
    package_data = _get_package_data(module.__name__)

    if package_type == 'plugin':
        qb_info.pop('config')
    else:
        # recipe
        qb_info.pop('entry_point')

    for k, v in package_data.items():
        if v is None:
            continue
        if k == 'name' and k in qb_info:
            # only use package name if name is not provided
            continue
        qb_info[k] = v

    metadata = MetaData.parse_obj(qb_info)

    return metadata


@dataclass
class _BaseClass:
    """Base class for Pollination dsl Function and DAG.

    Do not use this class directly.
    """
    _cached_queenbee = None
    _cached_outputs = None
    _cached_package = None
    _cached_inputs = None

    @property
    def queenbee(self):
        raise NotImplementedError

    @property
    def _outputs(self) -> NamedTuple:
        raise NotImplementedError

    @property
    def _inputs(self) -> NamedTuple:
        """Return inputs as a simple object with dot notation.

        Use this property to access the inputs when creating a DAG.

        The name starts with a _ not to conflict with a possible member of the class
        with the name inputs.
        """
        if self._cached_inputs:
            return self._cached_inputs
        cls_name = camel_to_snake(self.__class__.__name__)
        mapper = {
            inp.name.replace('-', '_'): {
                'name': inp.name.replace('-', '_'),
                'parent': cls_name,
                'value': inp
            } for inp in self.queenbee.inputs
        }

        inputs = namedtuple('Inputs', list(mapper.keys()))
        self._cached_inputs = inputs(*list(mapper.values()))

        return self._cached_inputs

    @property
    def _package(self) -> dict:
        """Queenbee package information.

        This information will only be available if the function is part of a Python
        package.
        """
        if self._cached_package:
            return self._cached_package

        module = import_module(self._python_package)
        assert hasattr(module, '__pollination__'), \
            'Failed to find __pollination__ info in __init__.py'
        package_data = _get_package_data(module.__name__)
        pollination_data = getattr(module, '__pollination__')
        for k, v in package_data.items():
            pollination_data[k] = v
        self._cached_package = pollination_data
        return self._cached_package

    @property
    def _python_package(self) -> str:
        """Python package information for this function.

        This information will only be available if the function is part of a Python
        package.
        """
        module = self.__module__
        if module.startswith('pollination'):
            return '.'.join(self.__module__.split('.')[:2])
        else:
            return self.__module__.split('.')[0]



from typing import NamedTuple
import re
import inspect
from collections import namedtuple
from typing import Dict, List

from queenbee.recipe.task import DAGTask, DAGTaskLoop
from queenbee.io.outputs.task import TaskReturn, TaskPathReturn
from queenbee.io.inputs.task import TaskArguments, TaskArgument, TaskPathArgument
from queenbee.base.parser import parse_double_quotes_vars

from .inputs import _InputBase as DAGInput
from ..common import camel_to_snake


def _validate_task_args(func) -> None:
    """Validate task arguments.

    This function ensures that the value for all the task arguments are provided.
    """
    func_args = inspect.getfullargspec(func)
    arg_names = func_args.args[1:]  # first arg is self
    defaults = func_args.defaults  # default values
    if len(arg_names) == 0:
        # no input arguments
        return
    # all args must have default values - raise an error if one is missing
    elif len(arg_names) != len(defaults):
        # number of arguments with missing values
        count = len(arg_names) - len(defaults)
        args = arg_names[: count]
        raise ValueError(
            f'Missing value for argument(s) in task "{func.__name__}" -> '
            f'{", ".join(args)}.\nAll the arguments should have an assigned value.'
        )


def _add_sub_path(arg: Dict, sub_paths: Dict) -> Dict:
    """Add the sub_path field to an argument."""
    name = arg['name'].replace('-', '_')
    if name in sub_paths:
        arg['sub_path'] = sub_paths[name]

    return arg


def _get_from(value, inputs_info) -> Dict:
    """Return a Queenbee From value dictionary.

    The output will be an InputReference, TaskReference or a ValueReference or a
    ValueListReference.

    Args:
        value: Reference object.
        inputs_info: DAG inputs info to get the information for DAGInput types.

    Returns:
        Dict - a Queenbee From value dictionary.
    """
    if isinstance(value, DAGInput):
        variable = inputs_info[id(value)]
        return {'type': value.reference_type, 'variable': variable}
    try:
        # only TaskReference has parent
        parent = value['parent']
    except TypeError:
        # value
        return {}
    else:
        # Task reference values and names are nested
        variable = value['name']
        value = value['value']
        return {'type': value.reference_type, 'name': parent, 'variable': variable}


def _get_task_arguments(func, inputs_info, sub_paths) -> List[TaskArguments]:
    """Get task input arguments in Python method as Queenbee task arguments."""
    task_args = []
    template = func.__task_template__
    func_args = inspect.getfullargspec(func)

    names = func_args.args[1:]  # first arg is self
    if not names:
        # no arguments
        return task_args

    values = func_args.defaults

    for name, value_info in zip(names, values):
        from_ = _get_from(value_info, inputs_info)
        arg_dict = {'name': name.replace('_', '-'), 'from': from_}
        arg_dict = _add_sub_path(arg_dict, sub_paths)
        if isinstance(value_info, DAGInput):
            if value_info.is_artifact:
                # file, folder, path
                arg = TaskPathArgument.parse_obj(arg_dict)
            else:
                arg = TaskArgument.parse_obj(arg_dict)
        elif from_ != {}:
            # task or value
            # tasks
            value = value_info['value']
            if value.is_artifact:
                arg = TaskPathArgument.parse_obj(arg_dict)
            else:
                # parameter
                arg = TaskArgument.parse_obj(arg_dict)
        else:
            # value reference
            func_input = getattr(template._inputs, name)['value']
            if func_input.is_artifact:

                arg_dict['from']['path'] = value_info

                if func_input.type == 'FunctionFolderInput':
                    # folder
                    arg_dict['from']['type'] = 'ValueFolderReference'
                else:
                    # file
                    arg_dict['from']['type'] = 'ValueFileReference'

                arg = TaskPathArgument.parse_obj(arg_dict)
            else:
                # value reference
                arg_dict['from']['value'] = value_info
                arg_dict['from']['type'] = 'ValueReference'
                arg = TaskArgument.parse_obj(arg_dict)

        task_args.append(arg)

    return task_args


def _get_task_loop(value, inputs_info) -> DAGTaskLoop:
    """Get TaskLoop as a queenbee TaskLoop."""
    if not value:
        return None
    from_ = _get_from(value, inputs_info)
    return DAGTaskLoop.parse_obj({'from': from_})


def _get_task_returns(func) -> NamedTuple:
    """Get task returns based on template outputs and returns."""
    template = func.__task_template__
    pattern = r'[\'\"]from[\'\"]\s*:\s*.*\._outputs\.(\S*)\s*[,}]'
    parent = func.__name__.replace('_', '-')
    src = inspect.getsource(func)
    # remove the last } which happens in case of parameters input. Somene who
    # knows regex better than I do should be able to fix this by changing the pattern
    # here is an example to recreate the issue.
    #   return [
    #       {'from': SplitGrid()._outputs.grids_list},
    #       {'from': SplitGrid()._outputs.output_folder, 'to': 'sub_grids'}
    #  ]
    matches = [match.replace('}', '') for match in re.findall(pattern, src)]
    mapper = {
        match: {
            'name': match.replace('_', '-'),
            'parent': parent,
            'value': getattr(template, match)
        } for match in matches
    }
    outputs = namedtuple('Outputs', list(mapper.keys()))
    return outputs(*list(mapper.values()))


def task(template, needs=None, loop=None, sub_folder: str = None, sub_paths: Dict = None,
         annotations: Dict = None):
    """Task decorator for methods in a DAG.

        @task(template=Template, ...)

        Args:
            template: A pollination-dsl Function or DAG.
            needs: A list of methods that this should be executed before this task.
                This task will only be executed after these tasks are executed
                successfully.
            loop: An iterator to loop over. The task will be executed for each item
                in this iterator.
            sub_folder: An optional sub_folder for execution of this task. If sub_folder
                is provided all the outputs from this task will be copied under the
                sub_folder. Sub_folder is usually used with loops and named based on
                the loop item to keep the results of each iteration in a separate folder.
            sub_paths: A dictionary to provide sub_path for input arguments. Sub_path is
                useful when a task only needs a file or a sub folder from the provided
                argument.
            annotations: An optional dictionary to annotate a task.

    """

    sub_paths = sub_paths or {}
    sub_paths = {key.replace('-', '_'): value for key, value in sub_paths.items()}

    def task_func(func):
        # add __decorator___ so I can find the tasks later in the class decorator
        func.__decorator__ = 'task'
        # set up task information
        # a template can only be from type dag or task
        assert hasattr(template, '__decorator__') and \
            getattr(template, '__decorator__') in {'function', 'dag'}, \
            f'Invalid input type for template: {template}\n' \
            'A template must be either a Function or a DAG.'
        func.__task_template__ = template()

        # technically a task should have _returns but I _felt_ it will be easier
        # for user to use _outputs here to keep it similar to Function and DAG.
        func._outputs = _get_task_returns(func)

        # check tasks that this task relies on
        func.__task_needs__ = needs or []
        for need in func.__task_needs__:
            assert hasattr(need, '__decorator__') and \
                getattr(need, '__decorator__') == 'task', \
                f'Invalid input type for needs: {need}. A task can only rely on ' \
                'another task in the same dag.'

        func.__task_loop__ = loop

        func.__task_annotations__ = annotations or {}
        func.__task_subfolder__ = sub_folder

        # double check task arguments
        _validate_task_args(func)

        # assign to_queenbee function to task as an inner function
        # this functions will be called when DAG object generates DAGTasks
        def to_queenbee(
                method, returns: List[Dict], dag_inputs: Dict[int, str], dag_package: str
                ):
            """Convert a task method to a Queenbee DAGTask."""
            name = method.__name__.replace('_', '-')
            tt = method.__task_template__
            if tt.__decorator__ == 'dag' and tt._package == dag_package:
                # template is another dag in the same plugin (AKA package)
                template = f'{camel_to_snake(tt.__class__.__name__)}'
            else:
                template = \
                    f'{tt._package["name"]}/{camel_to_snake(tt.__class__.__name__)}'

            task_needs = [
                need.__name__.replace('_', '-') for need in method.__task_needs__
            ]
            # TODO: validate arguments against needs to ensure all the task names are
            # included in needs.
            task_arguments = _get_task_arguments(method, dag_inputs, sub_paths)
            task_loop = _get_task_loop(method.__task_loop__, dag_inputs)
            task_returns = []
            for out in returns or []:
                from_ = out.get('from', None)
                from_ = from_['value'] if from_ else from_
                to_ = out.get('to', None)
                description = out.get('description', None)
                if from_.is_artifact:
                    assert to_ is not None, 'Missing \'to\' key for {from_.name}. ' \
                        'All file and folder returns must provide a target path using ' \
                        'the `to` key.'
                    if to_:
                        # find and replace referenced values
                        refs = parse_double_quotes_vars(to_)
                        for ref in refs:
                            to_ = to_.replace(
                                ref, ref.replace('self.', 'inputs.').replace('_', '-')
                            )
                    return_ = TaskPathReturn(
                        name=from_.name, description=description, path=to_
                    )
                else:
                    # returning a parameter
                    return_ = TaskReturn(name=from_.name, description=description)
                task_returns.append(return_)

            annotations = method.__task_annotations__
            sub_folder = method.__task_subfolder__
            return DAGTask(
                name=name, template=template, needs=task_needs, arguments=task_arguments,
                returns=task_returns, loop=task_loop, sub_folder=sub_folder,
                annotations=annotations
            )

        func.to_queenbee = to_queenbee

        return func

    return task_func



from dataclasses import dataclass
from typing import NamedTuple
import inspect
from collections import namedtuple

from queenbee.recipe.dag import DAG as QBDAG

from ..common import _BaseClass


@dataclass
class DAG(_BaseClass):
    """Baseclass for DSL DAG classes.

    Every Queenbee DAG must subclass from this class.

    Attributes:
        queenbee
        _dependencies
        _inputs
        _outputs
        _package
        _python_package

    """
    __decorator__ = 'dag'
    _cached_queenbee = None
    _cached_inputs = None
    _cached_outputs = None
    _cached_package = None

    @property
    def queenbee(self) -> QBDAG:
        """Convert this class to a Queenbee DAG."""
        # cache the DAG since it always stays the same for each instance
        if self._cached_queenbee:
            return self._cached_queenbee

        cls = self.__class__

        name = cls.__name__[0].lower() + \
            ''.join(['-' + x.lower() if x.isupper() else x for x in cls.__name__][1:])

        # create a mapper for inputs and use it to track back the name of the inputs
        # when creating a task reference by using the id of the assigned item.
        inputs_dict = {}
        for method_name, method in inspect.getmembers(cls):
            # try to get decorator
            qb_dec = getattr(method, '__decorator__', None)
            if qb_dec is None or qb_dec != 'input':
                continue
            inputs_dict[id(method)] = method_name.replace('_', '-')

        tasks = []
        inputs = []
        outputs = []

        for method_name, method in inspect.getmembers(cls):
            # try to get decorator
            qb_dec = getattr(method, '__decorator__', None)
            if qb_dec is None:
                continue
            if qb_dec == 'task':
                tasks.append(
                    method.to_queenbee(method, method(cls), inputs_dict, self._package)
                )
            elif qb_dec == 'input':
                inputs.append(method.to_queenbee(name=method_name))
            elif qb_dec == 'output':
                outputs.append(method.to_queenbee(name=method_name))
            else:
                raise ValueError(f'Unsupported __decorator__: {qb_dec}')

        self._cached_queenbee = QBDAG(
            name=name, inputs=inputs, tasks=tasks, outputs=outputs
        )
        return self._cached_queenbee

    @property
    def _outputs(self) -> NamedTuple:
        """Return dag outputs as a simple object with dot notation.

        Use this property to access the outputs when creating a DAG.

        The name starts with a _ not to conflict with a possible member of the class
        with the name outputs.
        """
        if self._cached_outputs:
            return self._cached_outputs

        mapper = {out.name: out for out in self.queenbee.outputs}
        outputs = namedtuple('Outputs', list(mapper.keys()))
        self._cached_outputs = outputs(*list(mapper.values()))

        return self._cached_outputs

    @property
    def _dependencies(self):
        """DAG dependencies.

        Dependencies are plugins or other recipes.
        """
        cls = self.__class__
        dag_package = self._package
        dependencies = {'plugin': [], 'recipe': [], 'dag': []}
        for method_name, method in inspect.getmembers(cls):
            # try to get decorator
            qb_dec = getattr(method, '__decorator__', None)
            if qb_dec == 'task':
                # get template
                tt = method.__task_template__
                if tt.__decorator__ == 'dag':
                    if tt._package == dag_package:
                        # it's a DAG in the same package.
                        if tt not in dependencies['dag']:
                            dependencies['dag'].append(tt)
                    else:
                        # dependency is another plugin
                        if tt._package not in dependencies['recipe']:
                            dependencies['recipe'].append(tt._package)
                elif tt.__decorator__ == 'function':
                    if tt._package not in dependencies['plugin']:
                        dependencies['plugin'].append(tt._package)

        return dependencies



import importlib
import pkgutil
import pathlib
import warnings
from typing import Union

from queenbee.plugin.plugin import Plugin, PluginConfig
from queenbee.recipe.recipe import Recipe, BakedRecipe, Dependency, DependencyKind
from queenbee.repository.package import PackageVersion
from queenbee.repository.index import RepositoryIndex
from queenbee.config import Config, RepositoryReference

from .function import Function
from .common import import_module, _get_meta_data, _get_package_readme, \
    get_requirement_version, name_to_pollination


def _init_repo() -> pathlib.Path:
    """Initiate a local Queenbee repository.

    This function is used by package function to start a local Queenbee repository
    if it doesn't exist. If the repository has already been created it will return
    the path to the repository.
    """
    HOME = pathlib.Path.home().as_posix()
    path = pathlib.Path(HOME, '.queenbee', 'pollination-dsl')
    path.mkdir(parents=True, exist_ok=True)

    index_file = path/'index.json'

    plugins_folder = path/'plugins'
    recipes_folder = path/'recipes'
    plugins_folder.mkdir(exist_ok=True)
    recipes_folder.mkdir(exist_ok=True)

    if not index_file.exists():
        index = RepositoryIndex.from_folder(path.as_posix())
        index.to_json(index_file.as_posix(), indent=2)

    return path


def _load_plugin(module) -> Plugin:
    """Load Queenbee plugin from Python package.

    Usually you should not be using this function directly. Use ``load`` function
    instead.

    args:
        module: Python module object for a Queenbee Plugin module.

    returns:
        Plugin - A Queenbee plugin
    """
    qb_info = module.__pollination__
    package_name = module.__name__
    # get metadata
    config = PluginConfig.parse_obj(qb_info['config'])
    metadata = _get_meta_data(module, 'plugin')

    folder = pathlib.Path(module.__file__).parent

    functions = []
    for (_, name, _) in pkgutil.iter_modules([folder]):
        module = importlib.import_module('.' + name, package_name)
        for attr in dir(module):
            loaded_attr = getattr(module, attr)
            if hasattr(loaded_attr, '__decorator__') and \
                    getattr(loaded_attr, '__decorator__') == 'function':
                if loaded_attr is Function:
                    continue
                functions.append(loaded_attr().queenbee)

    plugin = Plugin(config=config, metadata=metadata, functions=functions)
    return plugin


def package_recipe_dependencies(recipe: Recipe) -> None:
    """Try to load/package recipe dependencies from local registry.

    If the dependency is not available in local registry then it will try to package
    it from a local python installation. If that also fails it will try to pull it
    down using pip install.
    """
    for dep in recipe.dependencies:
        package(dep.name)


def _load_recipe(module, baked: bool = False) -> Union[BakedRecipe, Recipe]:
    # load entry-point DAG
    """Load Queenbee plugin from Python package.

    Usually you should not be using this function directly. Use ``load`` function
    instead.

    args:
        module: Python module object for a Queenbee Recipe.

    returns:
        Recipe - A Queenbee recipe. It will be a baked recipe if baked is set to True.
    """
    qb_info = module.__pollination__
    package_name = module.__name__

    main_dag_entry = qb_info.get('entry_point', None)
    assert main_dag_entry, \
        f'{package_name} __pollination__ info is missing the enetry_point key.'

    main_dag = main_dag_entry()

    # get metadata
    metadata = _get_meta_data(module, 'recipe')

    _dependencies = main_dag._dependencies
    # create a queenbee Recipe object

    # load dags
    qb_dag = main_dag.queenbee
    qb_dag.name = 'main'
    dags = [qb_dag] + [dag.queenbee for dag in _dependencies['dag']]

    # add dependencies
    repo = _init_repo()
    plugins = [
        Dependency(
            kind=DependencyKind.plugin, name=plugin['name'], tag=plugin['tag'],
            source=repo.as_uri()
        ) for plugin in _dependencies['plugin']
    ]
    recipes = [
        Dependency(
            kind=DependencyKind.recipe, name=recipe['name'], tag=recipe['tag'],
            source=repo.as_uri()
        ) for recipe in _dependencies['recipe']
    ]

    recipe = Recipe(metadata=metadata, dependencies=plugins + recipes, flow=dags)

    if baked:
        package_recipe_dependencies(recipe)
        rf = RepositoryReference(name='pollination-dsl', path='file:///' + repo.as_posix())
        config = Config(repositories=[rf])
        recipe = BakedRecipe.from_recipe(recipe=recipe, config=config)

    return recipe


def load(package_name: str, baked: bool = False) -> Union[Plugin, BakedRecipe, Recipe]:
    """Load Queenbee Plugin or Recipe from Python package.

        package_name: Python package name (e.g. honeybee-radiance-pollination)
        baked: A boolean value to indicate wether to return a Recipe or a BakedRecipe.
    """
    module = import_module(package_name)

    assert hasattr(module, '__pollination__'), \
        'Failed to find __pollination__ info in __init__.py'
    qb_info = getattr(module, '__pollination__')
    if 'config' in qb_info:
        # it's a plugin
        package = _load_plugin(module)
    else:
        # it's a recipe
        package = _load_recipe(module, baked)
        # try to update recipe tag based on requirements
        for dep in package.dependencies:
            name = name_to_pollination(dep.name)
            try:
                tag = get_requirement_version(package_name, name)
            except AssertionError:
                warnings.warn(
                    f'{package_name} has dependencies on {name} but it is not set as one '
                    'of the package dependencies in setup.py. Will use the version of the '
                    'currently installed version instead.'
                )
            else:
                dep.tag = tag

    return package


def package(package_name: str, readme: str = None) -> None:
    """Package a plugin or a recipe and add it to Queenbee local repository.

        Args:
            package_name: Python package name (e.g. honeybee-radiance-pollination)
            readme: Readme contents as a string.
    """
    # init a Queenbee package
    repository_path = _init_repo()
    index_path = repository_path/'index.json'
    qb_obj = load(package_name, baked=False)

    if isinstance(qb_obj, Recipe):
        qb_type = 'recipe'
        package_recipe_dependencies(qb_obj)
    elif isinstance(qb_obj, Plugin):
        qb_type = 'plugin'
    else:
        raise TypeError('Input package must be a Queenbee Plugin or a Queenbee Recipe.')

    # add to pollination-dsl repository
    try:
        plugin_version, file_object = PackageVersion.package_resource(
            qb_obj, readme=_get_package_readme(package_name)
        )
    except Exception as error:
        raise ValueError(f'Failed to package {package_name} {qb_type}\n {error}')

    file_path = repository_path/f'{qb_type}s'/plugin_version.url
    file_object.seek(0)
    file_path.write_bytes(file_object.read())

    # re-index the repository
    repo_index = RepositoryIndex.from_folder(repository_path)
    repo_index.to_json(index_path.as_posix(), indent=2)


def translate(
        package_name: str,
        target_folder: str,
        baked: bool = False,
        readme: str = None
        ) -> str:
    """Translate Python package to a Queenbee plugin or recipe.

    BakedRecipes are written to a single file.

    args:
        package_name: Python package name. The package must be installed
            in the environment that this command being executed.
        target_folder: Path to folder to write this plugin.
        baked: A boolean to write the Recipes as BackedRecipes. It will be ignored for
            plugins.
        readme: Readme contents as a string.

    returns:
        str -- path to the generated folder or file.
    """
    qb_object = load(package_name, baked=baked)
    if baked and isinstance(qb_object, BakedRecipe):
        recipe_file = pathlib.Path(target_folder, package_name + '.yaml')
        recipe_file.write_text(qb_object.yaml())
        return recipe_file.as_posix()
    else:
        qb_object.to_folder(
            folder_path=target_folder,
            readme_string=_get_package_readme(package_name)
        )
        return target_folder



"""All function related decorators and objects including inputs and outputs."""
from .inputs import Inputs  # expose for easy import
from .outputs import Outputs  # expose for easy import
from .base import Function, command
